/******************************************************************************
*                                                                             *
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <time.h>

typedef struct stats {
   char stn[4];   /* the 2 character title ie ST, AG */
   int  value;    /* rolled value */
   int  norm;     /* value bonus */
   int  race;     /* racial bonus */
   int  total;    /* bonus total */
} stats;

typedef struct abil {
   char title[20]; /* ability title ie. Soft Leather */
   int  rank;      /* no. of ranks ie. 1 */
   int  stat;      /* which stat to use on this ability - see chstat */
   int  stat_val;  /* ability bonus total */
   int  prof;      /* profession level bonus */
   int  item;
   int  tmp_dp;    /* temp. development points for current lvl */
   int  spec1;
   int  total;     /* total rank value */
   int  max_rank;  /* Max. no. of ranks */
   int  dev_rnk2;  /* no. of points to develop 2nd rank */
} abils;

typedef struct proftype {
   char profname[12]; /* Title of Profession */
   int  prim_abil;    /* which ability is the primary ability into chstat */
   int  dev_pts[10];  /* development points to spend per level */
   int  prof_bon[6];  /* profession bonuses. if < 50 then check into */
   int  prof_sbon[6]; /* prof_dp else it points directly into chabil */
   int  dev_wgt[8];   /* you can weight the development points, anything */
   int  realm;	      /* 0 = Ess, 1 = Cha, -1 = Either */
} proftype;           /* with a -1 will go into the pool */

typedef struct racedef {
   int  stat_bon[6];   /* racial stat bonuses */
   int  es_bon;        /* essence bonus */
   int  ch_bon;        /* channel bonus */
   int  po_bon;        /* poison bonus */
   int  di_bon;        /* disease bonus */
   int  initstats[24]; /* initial rankings in abilities */
   int  spell;         /* % chance of spell */
   int  add_lang;      /* additional languages */
   int  bgnd_pts;      /* background points */
   char race[20];      /* race name */
   char *spec;	       /* special message */
} racedef;

typedef struct statdef {
   int  start;         /* stat start range of bonus */
   int  end;           /* stat end range of bonus */
   int  bonus;         /* bonus value */
   int  pp;            /* no. of PPs */
} statdef;

typedef struct profess {
   int  abilptr[6];    /* point to the ability - see chabil structure */
   char category[40];  /* Skill category */
} profess;

typedef struct listrange {
   int  start;
   int  end;
   char desc[30];
} listrange;

struct chartype {
   char name[40];      /* character name */
   int  profession;    /* character profession */
   int  level;         /* current level */
   int  race;          /* racial type, links into race */
   int  pp;            /* no. PPs */
   int  type;          /* character type, 0 = character, 1 = NPC */
} chartype;

stats chstat[8] = {
  {"\0", 0, 0, 0, 0},
  {"ST", 0, 0, 0, 0},
  {"AG", 0, 0, 0, 0},
  {"CO", 0, 0, 0, 0},
  {"IG", 0, 0, 0, 0},
  {"IT", 0, 0, 0, 0},
  {"PR", 0, 0, 0, 0},
  {"AP", 0, 0, 0, 0}};

abils chabil[35] = {
  {"No Armor",		0, 2, 0, 0, 0, 0, 0, 0, 2, 1},
  {"Soft Leather",	0, 2, 0, 0, 0, 0, -15, 0, 3, 1},
  {"Rigid Leather",	0, 2, 0, 0, 0, 0, -30, 0, 5, 1},
  {"Chain",		0, 1, 0, 0, 0, 0, -45, 0, 7, 1},
  {"Plate",		0, 1, 0, 0, 0, 0, -60, 0, 9, 1},
  {"1H Edge",		0, 1, 0, 0, 0, 0, 0, 0, 20, 2},
  {"1H Conc",		0, 1, 0, 0, 0, 0, 0, 0, 20, 2},
  {"2Handed",		0, 1, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Thrown",		0, 2, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Missile",		0, 2, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Pole Arms",		0, 1, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Climb",		0, 2, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Ride",		0, 5, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Swim",		0, 2, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Track",		0, 4, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Ambush",		0, 0, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Stalk/Hide",	0, 6, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Pick Lock",		0, 4, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Disarm Trap",	0, 5, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Read Runes",	0, 4, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Use Items",		0, 5, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Directed Spells",	0, 2, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Body Devel.",	0, 3, 0, 0, 0, 0, 5, 0, 20, 2},
  {"Perception",	0, 5, 0, 0, 0, 0, 0, 0, 20, 2},
  {"Base Spells",	0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {"Leadership",	0, 6, 0, 0, 0, 0, 0, 0, 0, 0},
  {"Def. Bonus",	0, 2, 0, 0, 0, 0, 0, 0, 0, 0},
  {"Essence RR",	0, 4, 0, 0, 0, 0, 0, 0, 0, 0},
  {"Channel RR",	0, 5, 0, 0, 0, 0, 0, 0, 0, 0},
  {"Poison RR",		0, 3, 0, 0, 0, 0, 0, 0, 0, 0},
  {"Disease RR",	0, 3, 0, 0, 0, 0, 0, 0, 0, 0},
  {"Languages", 	0, 0, 0, 0, 0, 0, 0, 0, 300, 0},
  {"Spells", 		0, 0, 0, 0, 0, 0, 0, 0, 300, 0},
  {"Martial Arts",	0, 0, 0, 0, 0, 0, 0, 0, 15, 2},
  {"Martial Def",	0, 0, 0, 0, 0, 0, 0, 0, 15, 2}};

racedef race[22] = {
 {{-20,15,15,0,-5,-5},50,20,30,15, {1,0,0,0,0,0,0,0,2,2,0,2,0,0,0,0,5,1,1,0,0,0,2,4},0,3,5,"Hobbit", ""},
 {{5,0,10,0,-5,-5},20,0,5,5, {1,3,3,0,0,0,3,1,1,0,0,0,0,1,0,0,1,0,0,0,0,0,2,1},5,3,4,"Umli", ""},
 {{5,-5,15,0,-5,-5},40,0,10,10, {1,0,1,3,0,0,4,0,1,0,0,1,0,0,0,0,0,1,1,0,0,0,3,2},3,4,4,"Dwarf", ""},
 {{0,0,5,0,0,-5},20,0,0,0, {1,3,0,0,0,2,0,0,4,0,1,3,0,2,0,2,4,0,0,0,0,0,3,1},5,2,5,"Wose", ""}, 
 {{5,0,0,0,0,0},0,0,0,0, {1,0,0,0,0,0,0,1,1,0,2,2,0,2,0,0,4,0,0,0,0,0,3,1},3,3,5,"Beorning", "Wild animals wont attack"},
 {{5,0,0,0,0,0},0,0,0,0, {1,0,0,2,0,1,0,0,1,1,0,0,1,3,0,0,0,0,0,1,1,0,2,0},10,6,3,"Black Numenorean", ""},
 {{5,0,0,0,0,0},0,0,0,0, {1,0,2,0,0,2,0,0,1,1,0,0,0,5,0,0,0,0,0,0,0,0,2,0},5,5,5,"Corsair", ""},
 {{5,0,0,0,0,0},0,0,0,0, {1,0,1,0,0,0,1,0,1,1,1,0,2,1,0,0,0,0,0,0,0,0,1,0},10,5,5,"Dorwindrim", "+30 vs Poisons. +15 Acting/PublicSpeaking. +10 Rowing"},
 {{5,0,0,0,0,0},0,0,0,0, {1,0,1,0,0,0,1,0,2,1,2,5,0,1,0,0,2,0,0,0,0,0,3,1},2,2,4,"Dunlending", "+20 Climbing, PublicSpeaking, Acting & Acrobatic"},
 {{5,0,0,0,0,0},0,0,0,0, {1,0,0,0,0,1,0,0,1,2,2,0,5,0,0,0,0,0,0,0,0,0,2,1},2,2,4,"Easterling", ""},
 {{5,0,0,0,0,0},0,0,0,0, {1,1,0,0,0,1,0,0,1,0,2,0,7,0,0,0,0,0,0,0,0,0,2,1},2,3,5,"Haradrim", ""},
 {{5,0,0,0,0,0},0,0,0,0, {1,3,0,0,0,0,0,0,3,0,2,0,0,2,0,0,4,0,0,0,0,0,3,1},5,1,4,"Lossoth", ""},
 {{5,0,0,0,0,0},0,0,0,0, {1,0,1,2,0,2,0,0,0,1,1,0,8,1,0,0,0,0,0,0,0,0,2,1},3,4,5,"Rohirrim", "+20 OB on horseback"},
 {{5,0,0,0,0,0},0,0,0,0, {1,1,1,1,0,1,0,0,1,1,1,0,1,1,0,0,1,0,0,0,0,0,2,1},3,4,5,"Rural-Men", ""},
 {{5,0,0,0,0,0},0,0,0,0, {1,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,1,1,0,1,1},15,5,5,"Urban-Men", ""},
 {{5,0,0,0,0,0},0,0,0,0, {1,1,0,0,0,2,0,0,1,1,1,0,4,0,0,0,1,0,0,0,0,0,2,1},5,3,4,"Variag", ""},
 {{5,0,0,0,0,0},0,0,0,0, {1,1,0,0,0,1,0,0,1,1,1,3,0,1,0,0,4,0,0,0,0,0,2,1},3,2,5,"Woodmen", "+10 Foraging. +20 Climbing & Acrobatics"},
 {{5,0,10,0,0,5},0,0,5,5, {1,0,1,2,0,2,0,1,0,1,1,0,1,1,0,0,0,0,0,1,1,0,3,0},10,6,3,"Dunedain", ""}, 
 {{5,5,5,0,0,-5},0,0,5,50, {1,0,1,1,0,1,0,0,0,2,0,1,1,1,0,0,2,0,0,1,0,0,1,1},10,4,3,"Half-Elf", "+5 vs Cold. Vision"},
 {{0,10,0,0,5,5},0,0,10,100, {1,0,0,0,0,1,0,0,0,3,0,2,1,3,0,0,4,0,0,1,0,0,1,3},20,6,4,"Silvan Elf", "+10 vs Cold. Vision. +10 Bard Spells. +10 Trickery & Stalk/Hide"},
 {{0,10,5,0,5,10},0,0,10,100, {1,0,0,0,0,1,0,0,0,2,0,1,1,2,0,0,3,0,0,1,1,0,1,3},30,8,3,"Sindar Elf", "+15 vs Cold. Vision"},
 {{0,15,10,5,5,15},0,0,10,100, {1,0,0,0,0,1,0,0,0,1,0,0,1,2,0,0,2,0,0,2,1,0,1,3},40,10,2,"Noldor Elf", "+20 UseItems. +20 vs Cold. Vision. Hearing"}
};

statdef statval[13] = {
 {102,200,35,4},
 {101,101,30,3},
 {100,100,25,3},
 {98,99,20,2},
 {95,97,15,2},
 {90,94,10,1},
 {75,89,5,1},
 {25,74,0,0},
 {10,24,-5,0},
 {5,9,-10,0},
 {3,4,-15,0},
 {2,2,-20,0},
 {1,1,-25,0}
};

proftype profdev[13] = {
 {"Warrior", 1, {3,5,2,2,0,3,0,0,0,0},{2,1,72,-1,-1,-1},{1,3,2,0,0,0}, {3,1,4,5,0,2,0,0}, -1},  /* Warrior */
 {"Mage", 4, {0,0,2,0,5,1,2,5,0,0},{69,70,74,71,-1,-1},{2,2,2,3,0,0}, {0,0,4,0,1,3,-1,2}, 0},  /* Mage */
 {"Animist", 5, {1,1,2,1,2,1,2,5,0,0},{2,69,70,73,71,74},{1,1,1,1,2,2}, {6,2,3,7,5,4,-1,1}, 1},  /* Animist */
 {"Scout", 2, {1,3,3,5,0,2,1,0,0,0},{1,2,3,73,-1,-1},{1,1,2,3,0,0}, {3,2,5,1,0,4,6,0}, -1},  /* Scout */
 {"Ranger", 3, {2,3,4,2,0,2,1,1,0,0},{1,2,73,66,-1,-1},{2,3,2,2,0,0}, {4,1,3,6,0,2,0,5}, 1},  /* Ranger */
 {"Bard", 6, {0,2,2,2,3,1,3,2,0,0},{1,2,3,4,74,73},{1,1,1,1,1,1}, {0,2,3,7,1,6,4,5}, 0}, /* Bard */
 {"Rogue", 1, {2,4,3,3,1,2,0,0,0,0},{1,2,3,-1,-1,-1},{2,2,2,0,0,0}, {3,1,4,5,0,2,0,0}, -1},   /* Rogue */
 {"War. Monk", 1, {0,1,1,2,0,3,0,0,5,3},{2,1,72,-1,-1,-1},{1,3,2,0,0,0}, {3,1,4,5,0,2,0,0}, 0},   /* Warrior Monk */
 {"Monk", 2, {0,1,2,2,1,2,1,1,3,2},{2,1,72,-1,-1,-1},{1,3,2,0,0,0}, {3,1,4,5,0,2,0,0}, 0},  /* Monk */
 {"Mentalist", 5, {0,1,1,1,3,2,2,5,0,0},{2,69,70,73,71,74},{1,1,1,1,2,2}, {6,2,3,7,5,4,-1,1}, 0},  /* Mentalist */
 {"Conjurer", 5, {0,1,1,4,3,1,1,4,0,0},{4,3,73,74,-1,-1},{1,1,1,1,0,0}, {3,1,4,5,0,2,0,0}, -1},   /* Conjurer */
 {"Anim-Warr", 5, {1,3,2,1,2,1,1,4,0,0},{2,1,70,73,71,74},{1,1,1,1,1,1}, {6,2,3,7,5,4,-1,1}, 1},  /* Animist/Warrior */
 {"Dabbler", 4, {1,2,2,4,2,1,1,2,0,0},{1,2,3,74,-1,-1},{1,1,1,1,0,0}, {0,0,4,0,1,3,-1,2}, 0},  /* Mage/Thief */
};

profess prof_dp[11] = {
 {{0,1,2,3,4,-1},      "Movement & Meneuver"},
 {{5,6,7,8,9,10},      "Weapon"},
 {{11,12,13,14,-1,-1}, "General"},
 {{15,16,17,18,-1,-1}, "Subterfuge"},
 {{19,20,21,-1,-1,-1}, "Magical"},
 {{22,-1,-1,-1,-1,-1}, "Body Dev"},
 {{31,-1,-1,-1,-1,-1}, "Languages"},
 {{32,-1,-1,-1,-1,-1}, "Spells"},
 {{33,-1,-1,-1,-1,-1}, "Martial Arts"},
 {{34,-1,-1,-1,-1,-1}, "Martial Def"},
 {{-1,-1,-1,-1,-1,-1}, "Secondary Skills"},
};

listrange bgnditem0[] = {{0,60,"+10 Item/Skill or +1 Adder"}, {61,89,"Daily Spell Item"}, {90,200,"+15 Item/Skill or +2 Adder"}};
listrange bgnditem1[] = {{0,50,"+5 Primary Skill"},{51,55,"+15 Secondary Skill"},{56,60,"1 Animal Empathy (+25)"},{61,65,"100' Infravision"},{66,70,"+10 to 1 RR"},{71,75,"1 Spell List"},{76,80,"+10 to all MM's"},{81,85,"+10 Perception"},{86,90,"+5 DB/OB's"},{91,95,"+10 to Leadership/Influence"},{96,200,"+3 to each Hit roll"}};
listrange bgnditem2[] = {{0,2,"1gp"},{3,5,"2gp"},{6,15,"5gp"},{16,25,"10gp"},{26,35,"15gp"},{36,45,"20gp"},{46,55,"30gp"},{56,65,"35gp"},{66,70,"40gp"},{71,75,"50gp"},{76,80,"60gp"},{81,85,"70gp"},{86,90,"80gp"},{91,94,"100gp"},{95,97,"125gp"},{98,99,"150gp"},{100,200,"200gp"}};

char *skillhdr[] = {"Armor", "Weapons", "General", "Subterfuge", "Magic", "Hits","",""}; 

/******************************************************************************
*                                                                             *
******************************************************************************/
disp_char()
{
int i, lvl;
char pt_rnk[16];
 
  printf("\n");
  printf("Name  : %-40.40s\n", chartype.name);
  printf("Type  : %-12.12s\n", profdev[chartype.profession].profname);
  printf("Level : %2.2d\n", chartype.level);
  printf("Race  : %-20.20s\n", race[chartype.race].race);
  lvl=chartype.level;
  chartype.level = (chartype.level >0)?lvl:1;
  printf("PP    : %2.2d\n\n", chartype.pp*chartype.level);
  for (i=1;i<8;i++) 
    printf("%-3.3s    %+2.2d  %+2.2d  %+2.2d  %+2.2d\n",
      chstat[i].stn,chstat[i].value,chstat[i].norm,chstat[i].race,
      chstat[i].total);

  printf("\n");
  for (i=0;i<31;i++)
  {
    strncpy(pt_rnk, "               ", 15);
    strncpy(pt_rnk, "***************", chabil[i].max_rank);
    strncpy(pt_rnk, "XXXXXXXXXXXXXXX", chabil[i].rank);
    if (i==5 || i == 11 || i == 15 || i == 19 || i == 22 || i == 24 || i == 31)
       printf("\n");
    printf("%-16.16s  %-10.10s %-5.5s  %+2.2d  %+2.2d  %+2.2d  %+2.2d  %+2.2d\n",
     chabil[i].title, pt_rnk, &pt_rnk[10], chabil[i].stat_val, 
     (chabil[i].prof * chartype.level), chabil[i].item, chabil[i].spec1, 
      chabil[i].total);
  }
  for (i=31;i<33;i++)
  {
    printf("%-16.16s            %2.2d      %+2.2d  %+2.2d  %+2.2d  %+2.2d  %+2.2d\n",
     chabil[i].title, chabil[i].rank, chabil[i].stat_val, 
     (chabil[i].prof * chartype.level), chabil[i].item, chabil[i].spec1, 
      chabil[i].total);
  }
  printf("\n");
  for (i=33;i<35;i++)
  {
    strncpy(pt_rnk, "               ", 15);
    strncpy(pt_rnk, "***************", chabil[i].max_rank);
    strncpy(pt_rnk, "XXXXXXXXXXXXXXX", chabil[i].rank);
    printf("%-16.16s  %-10.10s %-5.5s  %+2.2d  %+2.2d  %+2.2d  %+2.2d  %+2.2d\n",
     chabil[i].title, pt_rnk, &pt_rnk[10], chabil[i].stat_val, 
     (chabil[i].prof * chartype.level), chabil[i].item, chabil[i].spec1, 
      chabil[i].total);
  }
  chartype.level=lvl;
}


/******************************************************************************
*                                                                             *
******************************************************************************/
get_stat(class, raceno)
int class, raceno;
{
int i, j, tmp=0;

  for (i=1; i < 7; i++)   /* set RACIAL stat bonuses */
  {
    chstat[i].race  = race[raceno].stat_bon[i-1];
  }
  while (tmp < 2)	  /* must have 2 scores > 90 */
  {
    tmp = 0;
    for (i=1; i < 8; i++)
    {
      while ((chstat[i].value = chrand()%100) < 30);
    }
    for (i=1; i < 7; i++)
    {
	tmp += (chstat[i].value > 90)? 1 : 0;
    }
  }

  for (i=1; i < 7; i++)
  {
    chstat[i].stn[2] = '\0';
    chstat[i].norm  = findnorm(chstat[i].value, &tmp);
    chstat[i].total = chstat[i].norm + chstat[i].race;
  }
 					 /* set AP = RND() + PR */
  chstat[7].value = chrand()%100 + chstat[6].total; 
  chstat[7].value = (chstat[7].value > 100)?100:chstat[7].value;
  
printf("Class = %d\n", class);
  if (class == 1 || class == 5 || class == 12)  /* get Essence PPs */
  {
     i = findnorm(chstat[4].value, &chartype.pp);
  }
  else if (class == 2 || class == 4 || class == 11)  /* get Channel PPs */
  {
     i = findnorm(chstat[5].value, &chartype.pp);
  }
  i = profdev[class].prim_abil;
  chstat[i].stn[2] = '*';
}

/******************************************************************************
*                                                                             *
******************************************************************************/
gen_char(class, raceno)
int class, raceno;
{
int i, j, tmp=0;

  if (class < 0 || class > 12 || raceno < 0 || raceno > 21)
  {
	printf("Bad Class or Race ... Choose again\n");
	return(-1);
  }

  get_stat(class, raceno);

  for (i=0; i < 24; i++)	/* set the initial RACIAL ranks */
  {
    if (chabil[i].stat > 0)
      chabil[i].stat_val = chstat[chabil[i].stat].total;
    chabil[i].rank = race[raceno].initstats[i];
  }
  for (i=23; i < 31; i++)
  {
    if (chabil[i].stat > 0)
      chabil[i].stat_val = chstat[chabil[i].stat].total;
  }
  chabil[27].spec1 = race[raceno].es_bon;  /* RACIAL essence bonnus */
  chabil[28].spec1 = race[raceno].ch_bon;  /* RACIAL channel bonnus */
  chabil[29].spec1 = race[raceno].po_bon;  /* RACIAL poison bonnus */
  chabil[30].spec1 = race[raceno].di_bon;  /* RACIAL disease bonnus */
  chabil[31].rank = race[raceno].add_lang; /* RACIAL additional langs */

  for (i=0;i<6;i++)		/* now do the Profession bonuses */
  {
     if (profdev[class].prof_bon[i] < 0)
       break;
     else 
     {
       if (profdev[class].prof_bon[i] > 50)
          chabil[profdev[class].prof_bon[i]-50].prof 
                = profdev[class].prof_sbon[i];
       else  
         for (j=0;j<6;j++)
         {
            if (prof_dp[profdev[class].prof_bon[i]].abilptr[j] < 0)
               break;
            else
              chabil[prof_dp[profdev[class].prof_bon[i]].abilptr[j]].prof 
                   = profdev[class].prof_sbon[i];
         }
      }
   }
   for (i=0; i < 24; i++)
   {
    tmp = (chabil[i].rank<11)?(chabil[i].rank*5):((chabil[i].rank-10)*2+50);
    chabil[i].total = chabil[i].stat_val + tmp +
        (chartype.level * chabil[i].prof) + chabil[i].item + chabil[i].spec1;
    chabil[i].total += ((chabil[i].rank==0)?-25:0);
   }
   for (i=24; i < 31; i++)
   {
    chabil[i].total = chabil[i].stat_val + 
        (chartype.level * chabil[i].prof) + chabil[i].item + chabil[i].spec1;
   }
   if (race[raceno].spell <= chrand()%100)
   {
      chabil[32].rank += 5;
   }
   for (i=33; i < 35; i++)
   {
    chabil[i].rank += chabil[i].tmp_dp;
    tmp = (chabil[i].rank<11)?(chabil[i].rank*5):((chabil[i].rank-10)*2+50);
    chabil[i].total = chabil[i].stat_val + tmp + 
        (chartype.level * chabil[i].prof) + chabil[i].item + chabil[i].spec1;
    chabil[i].total += ((chabil[i].rank==0)?-25:0);
    chabil[i].tmp_dp = 0;
   }
}


/******************************************************************************
*                                                                             *
******************************************************************************/
loadchar(fn)
char fn[15];
{
int fp, i, e;

	if ((fp = open(fn, O_RDWR)) == -1)
	{
	   printf("Can't open %s.\n", fn);
	   return(0);
	}
	if ((i=read(fp, &chartype.name[0], sizeof(chartype))) 
			!= sizeof(chartype))
	{
	   printf("Error during read %d .....\n", i);
	   close(fp);
	   return(0);
	}
	for (i=0; i < 8; i++)
	{
	   if ((e=read(fp, &chstat[i].stn[0], sizeof(stats))) != sizeof(stats))
	   {
	      printf("Error during read %d .....\n", e);
	      close(fp);
	      return(0);
	   }
	}
	for (i=0; i < 35; i++)
	{
	   if ((e=read(fp, &chabil[i].title[0], sizeof(abils))) != 
			sizeof(abils))
   	   {
	      printf("Error during read %d .....\n", e);
	      close(fp);
	      return(0);
	   }
	}
	close(fp);
}

/******************************************************************************
*                                                                             *
******************************************************************************/
savasc(fn)
char fn[15];
{
FILE *fp;
char pt_rnk[26];
int i, e=0, lvl, hdr=0;

  if ((fp = fopen(fn, "w+")) == NULL)
  {
     printf("Can't save to %s.\n", fn);
     return(0);
  }

  lvl=chartype.level;
  chartype.level = (chartype.level >0)?lvl:1;
  fprintf(fp, "Name  : %-40.40s\n", chartype.name);
  fprintf(fp, "Type  : %-12.12s\n", profdev[chartype.profession].profname);
  fprintf(fp, "Level : %2.2d\n", chartype.level);
  i = profdev[chartype.profession].realm;
  if (i == -1) strcpy(pt_rnk, "(Either)");
  if (i == 0) strcpy(pt_rnk, "Essence");
  if (i == 1) strcpy(pt_rnk, "Channel");
  fprintf(fp, "Race  : %-s (%s)\n", race[chartype.race].race, race[chartype.race].spec);
  fprintf(fp, "PP    : %2.2d\n", chartype.pp*chartype.level);
  fprintf(fp, "Realm : %s\n\n", pt_rnk);
  for (i=1;i<7;i++) 
    fprintf(fp, "%-3.3s    %+2.2d  %+2.2d  %+2.2d  %+2.2d\n",
      chstat[i].stn,chstat[i].value,chstat[i].norm,chstat[i].race,
      chstat[i].total);
  fprintf(fp, "%-3.3s    %+2.2d\n", chstat[i].stn,chstat[i].value);

  fprintf(fp, "\n");
  fprintf(fp, "  Skill             Ranks Allocated        RV   PB   IB   SB   Tot    DP\n");
  fprintf(fp, "     %s\t\t\t\t\t\t\t\t%d\n", skillhdr[hdr++], profdev[chartype.profession].dev_pts[e++]);
  for (i=0;i<31;i++)
  {
    strncpy(pt_rnk, "                         ", 25);
    strncpy(pt_rnk, "*************************", chabil[i].max_rank);
    strncpy(pt_rnk, "XXXXXXXXXXXXXXXXXXXXXXXXX", chabil[i].rank);
    if (i==5 || i == 11 || i == 15 || i == 19 || i == 22) {
/*       fprintf(fp, "\t\t\t     ***** *****\n");  */
         fprintf(fp, "     %s\t\t\t\t\t\t\t\t%d\n", skillhdr[hdr++], profdev[chartype.profession].dev_pts[e++]);
    }
    if (i==31 || i==24 || i==23)
       fprintf(fp, "\n");
    fprintf(fp, "%-16.16s  %-10.10s %-5.5s %-5.5s  %+2.2d  %+2.2d  %+2.2d  %+2.2d  %+2.2d\n",
     chabil[i].title, pt_rnk, &pt_rnk[10], &pt_rnk[15], chabil[i].stat_val, 
     (chabil[i].prof * chartype.level), chabil[i].item, chabil[i].spec1, 
      chabil[i].total);
  }
  for (i=31;i<35;i++)
  {
    fprintf(fp, "\n");
    fprintf(fp, "%-16.16s            %2.2d            %+2.2d  %+2.2d  %+2.2d  %+2.2d  %+2.2d",
     chabil[i].title, chabil[i].rank, chabil[i].stat_val, 
     (chabil[i].prof * chartype.level), chabil[i].item, chabil[i].spec1, 
      chabil[i].total);
    fprintf(fp, "     %d\n", profdev[chartype.profession].dev_pts[e++]);
  }
  fprintf(fp, "\nChance of SpellList   : %d %%\n", race[chartype.race].spell);
  fprintf(fp, "Additional Languages  : %d Ranks\n", race[chartype.race].add_lang);
  fprintf(fp, "Background Pts        : %d\n", race[chartype.race].bgnd_pts);

  chartype.level=lvl;
  fflush(fp);
  close(fp);
}

/******************************************************************************
*                                                                             *
******************************************************************************/
savbin(fn)
char fn[15];
{
int fp, i, e;

	if ((fp = open(fn, O_CREAT | O_TRUNC | O_RDWR, 0644)) == -1)
	{
	   printf("Can't save to %s.\n", fn);
	   return(0);
	}
	if ((i=write(fp, &chartype.name[0], sizeof(chartype))) 
			!= sizeof(chartype))
	{
	   printf("Error during writing %d .....\n", i);
	   close(fp);
	   return(0);
	}
	for (i=0; i < 8; i++)
	{
	   if ((e=write(fp, &chstat[i].stn[0], sizeof(stats))) != sizeof(stats))
	   {
	      printf("Error during writing %d .....\n", e);
	      close(fp);
	      return(0);
	   }
	}
	for (i=0; i < 35; i++)
	{
	   if ((e=write(fp, &chabil[i].title[0], sizeof(abils))) != 
			sizeof(abils))
   	   {
	      printf("Error during writing %d .....\n", e);
	      close(fp);
	      return(0);
	   }
	}
	close(fp);
}

/******************************************************************************
*                                                                             *
******************************************************************************/
devel_npchar(raceno, class, cur_lvl, new_lvl)
int raceno, class, cur_lvl, new_lvl;
{
int i, j, k, tmp, maxrank=10, dp, cnt=0, maxt, maxv, maxtptr, pool=0;

  for (i=0;i<10;i++)		/* now do the developing */
  {
/* printf("I = %d   DP = %d\n",i, profdev[class].dev_pts[i]); */
     if ((dp=profdev[class].dev_pts[i]) != 0)
     {
	 if (i > 0 && chrand()%100 < 15)  /* spend 1 DP on the lowest score */
	 {	 			  /* just for varity */
	    maxt = 200;
	    for (j=0;j<6;j++)
                if (chabil[prof_dp[i].abilptr[j]].total < maxt
                 && prof_dp[i].abilptr[j] > 0)
		{
                   maxt = chabil[prof_dp[i].abilptr[j]].total;
	  	   maxtptr = j;
		}
	    dp--;
            j = prof_dp[i].abilptr[maxtptr];
/* printf("Update Lowest score = %d\n", j); */
            chabil[j].rank++;
    	    tmp = (chabil[j].rank<11)?(chabil[j].rank*5):
				   ((chabil[j].rank-10)*2+50);
    	    chabil[j].total = chabil[j].stat_val + tmp + 
            	(chartype.level * chabil[j].prof) + chabil[j].item + 
		 chabil[j].spec1;
    	    chabil[j].total += ((chabil[j].rank==0)?-25:0);
	 }
	 maxt = -100;
	 maxv = 300;
	 maxtptr = -1;
	 maxrank = 10;
	 if (i == 0 && dp > 0)
	 {
	   if (class == 2 || class == 6)
	   {
	      chabil[0].tmp_dp += (dp>1)?2:1;
	      dp -= (dp>1)?2:1;
	   }
	   else if (class == 3 || class == 5)
	   {
	      chabil[2].tmp_dp += (dp>1)?2:1;
	      dp -= (dp>1)?2:1;
	      if (dp > 0)
	         chabil[0].tmp_dp += (dp>1)?2:1;
	   }
	   else
	   {
	      if (chabil[3].total > chabil[4].total)
	         chabil[3].tmp_dp += (dp>1)?2:1;
	      else
	         chabil[4].tmp_dp += (dp>1)?2:1;
	      dp -= (dp>1)?2:1;
	      if (dp > 0)
	         chabil[0].tmp_dp += (dp>1)?2:1;
	   }
	   for (j=0;j<5;j++)
	   {
	      if ((chabil[j].rank + chabil[j].tmp_dp) > chabil[j].max_rank)
	      {
	        tmp = (chabil[j].rank + chabil[j].tmp_dp) - chabil[j].max_rank;
		if (tmp > 0)
		{
	          chabil[j].tmp_dp -= tmp;
		  dp += tmp;
		}
	      }
	   }
	   if (dp > 0)
	   {
		pool += dp;
		dp = 0;
	   }
	 }
	 while (dp > 0)
	 {
	   maxt = -100;
           for (j=0;j<6;j++)
           {
/* printf("J = %d   ABILPTR = %d\n",j, prof_dp[i].abilptr[j]); */
              if (prof_dp[i].abilptr[j] < 0)
                 break;
              else
                if (chabil[prof_dp[i].abilptr[j]].total > maxt
                 && chabil[prof_dp[i].abilptr[j]].total < maxv)
	        {
                   maxt = chabil[prof_dp[i].abilptr[j]].total;
	  	   maxtptr = j;
	        }
/* printf("T = %d    Ptr = %d  Val = %d  DP = %d  VPtr = %d\n",maxt, maxtptr, chabil[prof_dp[i].abilptr[j]].total, dp, prof_dp[i].abilptr[j]);  */
           }
	   cnt++;
	   maxv = maxt;
	   if (dp > 1)
	   {
		if (dp >= (1 + chabil[prof_dp[i].abilptr[maxtptr]].dev_rnk2))
		{
		  tmp = 1 + chabil[prof_dp[i].abilptr[maxtptr]].dev_rnk2;
		}
		else
		{
		  tmp = 1;
		}
           }
	   else
	   {
	     tmp = 1;
	   }
	   if (prof_dp[i].abilptr[maxtptr] == 32   /* handle spells/langs */
	    || prof_dp[i].abilptr[maxtptr] == 31)  /* differently */
  	   {
	     chabil[prof_dp[i].abilptr[maxtptr]].rank += dp;
	     break;
  	   }
	      
	   k = chabil[prof_dp[i].abilptr[maxtptr]].rank;
	   if (((k + tmp) <= chabil[prof_dp[i].abilptr[maxtptr]].max_rank)
	     && ((k + tmp) <= maxrank))
	   {
	     chabil[prof_dp[i].abilptr[maxtptr]].tmp_dp += (tmp>1)?2:1;
	     dp -= tmp;
	   }
	   else if (((k + 1) <= chabil[prof_dp[i].abilptr[maxtptr]].max_rank)
	     && ((k + 1) <= 10))
	   {
	     chabil[prof_dp[i].abilptr[maxtptr]].tmp_dp += 1;
	     dp -= 1;
	   }
	   else
	   {
	     dp += chabil[prof_dp[i].abilptr[maxtptr]].tmp_dp;
	     chabil[prof_dp[i].abilptr[maxtptr]].tmp_dp = 0;
	   }
	   if (cnt > 36)
	   {
	      pool += dp;
	      break;
	   }
	   if (cnt%6 == 0 && dp > 0) 
	   {
	      maxrank += 2;
	      maxrank = (maxrank > 15)?15:maxrank;
	   }
	 }
      }
   }

   while (pool > 0)
   {
	if ((chabil[23].rank + chabil[23].tmp_dp + (pool > 2)?2:1) 
			<= chabil[23].max_rank)
	{
	  chabil[23].tmp_dp += (pool > 2)?2:1;   /* any spare ? put onto Perc */
	  pool -= (pool > 2)?3:1;
	  pool = 0;
	}
   }
   for (i=0; i < 24; i++)
   {
    chabil[i].rank += chabil[i].tmp_dp;
    tmp = (chabil[i].rank<11)?(chabil[i].rank*5):((chabil[i].rank-10)*2+50);
    chabil[i].total = chabil[i].stat_val + tmp + 
        (chartype.level * chabil[i].prof) + chabil[i].item + chabil[i].spec1;
    chabil[i].total += ((chabil[i].rank==0)?-25:0);
    chabil[i].tmp_dp = 0;
   }
   for (i=24; i < 31; i++)
   {
    chabil[i].total = chabil[i].stat_val + 
        (chartype.level * chabil[i].prof) + chabil[i].item + chabil[i].spec1;
   }
   for (i=33; i < 35; i++)
   {
    chabil[i].rank += chabil[i].tmp_dp;
    tmp = (chabil[i].rank<11)?(chabil[i].rank*5):((chabil[i].rank-10)*2+50);
    chabil[i].total = chabil[i].stat_val + tmp + 
        (chartype.level * chabil[i].prof) + chabil[i].item + chabil[i].spec1;
    chabil[i].total += ((chabil[i].rank==0)?-25:0);
    chabil[i].tmp_dp = 0;
   }
}

/******************************************************************************
*                                                                             *
******************************************************************************/
devel_char(raceno, class, cur_lvl, new_lvl)
int raceno, class, cur_lvl, new_lvl;
{
int i, j, k, tmp, pool=0, dp, abil[6]; 

   for (j=0;j<8;j++)
   {
      if (profdev[class].dev_wgt[j] == -1)
         pool += profdev[class].dev_pts[j];
   }

   for (j=0;j<8;j++)
   {
      if (profdev[class].dev_pts[j] > 0)
         for (i=0;i<8;i++)
         {
       
         }
   }

   for (i=0; i < 24; i++)
   {
    tmp = (chabil[i].rank<11)?(chabil[i].rank*5):((chabil[i].rank-10)*2+50);
    chabil[i].total = chabil[i].stat_val + tmp + 
        (chartype.level * chabil[i].prof) + chabil[i].item + chabil[i].spec1;
    chabil[i].total += ((chabil[i].rank==0)?-25:0);
   }
   for (i=24; i < 31; i++)
   {
    chabil[i].total = chabil[i].stat_val + 
        (chartype.level * chabil[i].prof) + chabil[i].item + chabil[i].spec1;
   }
}

/******************************************************************************
*                                                                             *
******************************************************************************/
void bgnd_pts(val)
int val;
{
int i;
listrange *p;

   if (val == 0) {
	p = bgnditem0;
   }
   else if (val == 1) {
	p = bgnditem1;
   }
   else if (val == 2) {
	p = bgnditem2;
   }
   i = chrand()%100;
   while (1) {
/* printf("%d:%d - %s (%d)\n", p->start, p->end, p->desc, i); */
        if (p->start <= i && p->end >= i) {
	   printf("%s (%d)\n", p->desc, i);
	   break;
	}
	p++;
   }
}

/******************************************************************************
*                                                                             *
******************************************************************************/
int findnorm(val, pp)
int val, *pp;
{
int i;

   for (i=0; i < 13; i++)
     if (val >= statval[i].start && val <= statval[i].end)
     {
	*pp = statval[i].pp;
        return(statval[i].bonus);
     }

   return(0);
}


/******************************************************************************
*                                                                             *
******************************************************************************/
int chrand()
{ 
int rnd;

   srand(rand());
   return(rand());
}


/******************************************************************************
*                                                                             *
******************************************************************************/
cli()
{
char inp[40],ginp[40],tstr[40];
int raceno, class, i, race, type, lvl, tmp;

   while (1)
   {
      printf("=> ");
      gets(ginp);

      tmp = i = 0;
      while(ginp[tmp] != '\0')
      {
	if (ginp[tmp] != ' ')
	  inp[i++] = tolower(ginp[tmp]);
	tmp++;
      }
      inp[i] = '\0';

      switch(inp[0])
      {
          case 'h': 
          case '?': help();
		    break;

          case 'b': switch(inp[1])
		    {
			case 'g': bgnd_pts(atoi(&inp[2]));
				  break;
	   		default: printf("Bad command ... use ? for help\n");
		    		  break;
		    }
		    break;

          case 'q': exit(0);
		    break;

	  case 'd': switch(inp[1])
		    {
			case 'c': disp_char();
				  break;
			case 's': for (i=1;i<8;i++) 
                                    printf(
				    "%-3.3s    %+2.2d  %+2.2d  %+2.2d  %+2.2d\n",
                                      chstat[i].stn,chstat[i].value,
			 	      chstat[i].norm,chstat[i].race,
      				      chstat[i].total);
				  printf("\n");
				  break;

			case 'v': lvl = inp[2] - 48;
				  for (i=chartype.level;i<=lvl;i++)
				    if (chartype.type == 0)
				       devel_char(chartype.race, 
				  	  chartype.profession,
					  i, lvl);
				    else
				       devel_npchar(chartype.race, 
					  chartype.profession,
					  i, lvl);
				  chartype.level = lvl;
				  break;
	   		default: printf("Bad command ... use ? for help\n");
		    		  break;
		     }
		    break;

          case 's': switch(inp[1])
		    {
			case 'a': strcpy(tstr, &inp[2]);
				  savasc(tstr);
			          break;

			case 'b': strcpy(tstr, &inp[2]);
				  savbin(tstr);
			          break;
	   		default: printf("Bad command ... use ? for help\n");
		    		  break;
		     }
		    break;

	  case 'e': switch(inp[1])
                    {
                        case 's': for (i=1;i<8;i++)
                                  {
				     printf("%-3.3s -> ",chstat[i].stn);
				     gets(tstr);
				     chstat[i].value = (int )atoi(tstr);
				     chstat[i].norm = findnorm(chstat[i].value,
								&tmp);
				     chstat[i].total = chstat[i].norm + 
                                                       chstat[i].race;
				  }
				  i = profdev[class].prim_abil;
				  if (chstat[i].value < 90) {
				    chstat[i].value = 90;
				    chstat[i].norm = 10;
				    chstat[i].total = 10 + chstat[i].race;
				  }
  				  for (i=0; i < 31; i++)
  				  {
    	  			    if (chabil[i].stat > 0)
      				       chabil[i].stat_val = 
				  	  chstat[chabil[i].stat].total;
  				  }
   				  for (i=0; i < 24; i++)
   				  {
    				    tmp = 
					(chabil[i].rank<11)?(chabil[i].rank*5):
						((chabil[i].rank-10)*2+50);
    				    chabil[i].total = chabil[i].stat_val + tmp 
					  + (chartype.level * chabil[i].prof) + 
				          chabil[i].item + chabil[i].spec1;
    				    chabil[i].total += 
					((chabil[i].rank==0)?-25:0);
   				  }
   				  for (i=24; i < 31; i++)
   				  {
				      chabil[i].total = chabil[i].stat_val + 
				          (chartype.level * chabil[i].prof) + chabil[i].item + chabil[i].spec1;
   				  }
  				  if (class == 1 || class == 5 || class == 12)
  				  {		   /* get Essence PPs */
     				    i = findnorm(chstat[4].value, &chartype.pp);
  				  }
  				  if (class == 2 || class == 4 || class == 11)
  				  {		 /* get Channel PPs */
     				    i = findnorm(chstat[5].value, &chartype.pp);
  				  }
                                  break;

                        case 'n': printf("Name > ");
                                  gets(chartype.name);
                                  break;
	   		default: printf("Bad command ... use ? for help\n");
		    		  break;
                     }
		    break;

	   case 'g': switch(inp[1])
		     {
			case 'c': class = inp[2] - 97;
  			          raceno = inp[3] - 97;
  			          lvl = atoi(&inp[4]);
				  loadchar("empty.chr");
				  if (gen_char(class,raceno) == -1)
					break;
				  chartype.type = 0;
				  chartype.level = lvl;
				  chartype.race = raceno;
				  chartype.profession = class;
   				  for (i=0;i<lvl;i++)
     				    devel_char(raceno, class, i, lvl);
			          break;

			case 'n': class = inp[2] - 97;
  			          raceno = inp[3] - 97;
  			          lvl = atoi(&inp[4]);
				  loadchar("empty.chr");
				  if (gen_char(class,raceno) == -1)
					break;
				  chartype.profession = class;
				  chartype.type = 1;
				  chartype.level = lvl;
				  chartype.race = raceno;
   				  for (i=0;i<lvl;i++)
     				    devel_npchar(raceno, class, i, lvl);
				  i = profdev[class].prim_abil;
				  if (chstat[i].value < 90)  {
				    chstat[i].value = 90;
				    chstat[i].norm = 10;
				    chstat[i].total = 10 + chstat[i].race;
				  }
			          break;
	   		default: printf("Bad command ... use ? for help\n");
		    		  break;
		     }
		    break;

	   case 'l': switch(inp[1])
		     {
			case 'd': strcpy(tstr, &inp[2]);
				  loadchar(tstr);
				  break;
		     }
		    break;

	   default: printf("Bad command ... use ? for help\n");
		    break;
      }
   }
}

/******************************************************************************
*                                                                             *
******************************************************************************/
help()
{
  printf("Select Character Type\n");
  printf("A - Warrior  B - Mage    C - Animist    D - Scout   E - Ranger\n");
  printf("F - Bard     G - Rogue   H - War. Monk  I - Monk    J - Mentalist\n");
  printf("K - Conjurer L - Anim-Wa M - Dabbler\n");
  printf("a - Hobbit          g - Corsairs     m - Rohirrim   s - Half-Elf\n");
  printf("b - Umli            h - Dorwindrim   n - Rural-Men  t - Silvan Elf\n");
  printf("c - Dwarf           i - Dunlendings  o - Urban-Men  u - Sindar Elf\n");
  printf("d - Wose            j - Easterlings  p - Variags    v - Noldor Elf\n");
  printf("e - Beornings       k - Haradrim     q - Woodmen\n");
  printf("f - B. Numenoreans  l - Lossoth      r - Dunedain\n\n");
/*  printf(" gc XYZ      - generate character, profession X, Race Y, Level Z\n"); */
  printf(" gn XYZ      - generate npc, profession X, Race Y, Level Z\n");
  printf(" sa filename - save in ascii format into file filename\n");
  printf(" sb filename - save in binary format into file filename\n");
  printf(" ld filename - load character from file, must be in binary format\n");
  printf(" dv lvl      - develop current character to level lvl\n");
  printf(" dc          - display current character on screen\n");
  printf(" ds          - display current character statistics on screen\n");
  printf(" es          - enter stats\n");
  printf(" en          - enter name\n");
  printf(" bg X        - use background pts, X = 0(Items), 1(Spec. Abil), 2(Money)\n");
  printf(" ?,h         - help, this message\n");
  printf(" qu          - quit/exit\n\n");
}

/******************************************************************************
*                                                                             *
******************************************************************************/
main()
{
#ifdef __GO32__
unsigned int ltime;
#else
long ltime;
#endif
struct tm *ntime;

   time(&ltime);
   ntime = gmtime(&ltime);
   srand(ntime->tm_sec+ntime->tm_min);
   cli();
}
